@@ -13,6 +13,8 @@
  * for 0.87: output "charset iso-8859-x" so that loadkeys
  *      can handle the output of dumpkeys -c
  * for 0.88: handle sparse keymaps
+ *
+ * for 0.94: support alt_is_meta
  */
 #include <stdio.h>
 #include <ctype.h>
@@ -34,11 +36,13 @@
 #define MAX_NR_KEYMAPS NR_KEYMAPS
 #endif
 
-#define VERSION "0.91"
+#define VERSION "0.94"
 extern int getfd();
 static int fd;
 
-char keymap_exists[MAX_NR_KEYMAPS];
+static int verbose;
+
+int keymap_index[MAX_NR_KEYMAPS];		/* inverse of good_keymap */
 int good_keymap[MAX_NR_KEYMAPS], keymapnr, allocct;
 
 void get_keymaps(void) {
@@ -56,12 +60,12 @@ void get_keymaps(void) {
 		exit(1);
 	    }
 	    if (!j && ke.kb_value != K_NOSUCHMAP) {
-		keymap_exists[i] = 1;
+		keymap_index[i] = keymapnr;
 		good_keymap[keymapnr++] = i;
 		if (ke.kb_value == K_ALLOCATED)
 		  allocct++;
 	    } else {
-		keymap_exists[i] = 0;
+		keymap_index[i] = -1;
 	    }
 	}
 	if (keymapnr == 0) {
@@ -123,15 +127,14 @@ void print_keysym(int code, char numeric) {
 	        t = KT_LATIN;
 		    printf("+");
 	}
-	if (
-		!numeric &&
-		t < syms_size &&
-		v < syms[t].size &&
-		(p = syms[t].table[v])[0]
-	)
-		printf("%-16s", p);
+	if (!numeric && t < syms_size && v < syms[t].size &&
+		 (p = syms[t].table[v])[0])
+	     printf("%-16s", p);
+	else if (!numeric && t == KT_META && v < 128 && v < syms[0].size &&
+		 (p = syms[0].table[v])[0])
+	     printf("Meta_%-11s", p);
 	else
-		printf("0x%04x          ", code);
+	     printf("0x%04x          ", code);
 }
 
 char valid_type(int t) {
@@ -166,6 +169,53 @@ u_char maximum_val(int t) {
 #define NR_TYPES 13
 int maxval[NR_TYPES];
 
+#ifdef KDGKBDIACR
+/* isgraph() does not know about iso-8859; printing the character
+   unescaped makes the output easier to check. Maybe this should
+   be an option. Use locale? */
+void
+outchar (unsigned char c) {
+        printf("'");
+        printf((c == '\'' || c == '\\') ? "\\%c"
+	       : (isgraph(c) || c == ' ' || c >= 0200) ? "%c"
+	       : "\\%03o", c);
+	printf("'");
+}
+
+struct kbdiacrs kd;
+
+void get_diacs(void) {
+	static int got_diacs = 0;
+
+	if(!got_diacs && ioctl(fd, KDGKBDIACR, (unsigned long)&kd)) {
+	    perror("");
+	    fprintf(stderr, "KDGKBDIACR failed\n");
+	    exit(1);
+	}
+	got_diacs = 1;
+}
+
+int nr_of_diacs(void) {
+	get_diacs();
+	return kd.kb_cnt;
+}
+
+void dump_diacs(void) {
+	int i;
+
+	get_diacs();
+	for (i = 0; i < kd.kb_cnt; i++) {
+	        printf("compose ");
+		outchar(kd.kbdiacr[i].diacr);
+		printf(" ");
+		outchar(kd.kbdiacr[i].base);
+		printf(" to ");
+		outchar(kd.kbdiacr[i].result);
+		printf("\n");
+	}
+}
+#endif        
+
 void show_short_info(void) {
 	int i;
 
@@ -183,9 +233,15 @@ void show_short_info(void) {
 	    maxval[i] = maximum_val(i);
 	    printf("	0x%04x - 0x%04x\n", K(i, 0), K(i, maxval[i]));
 	}
-	printf("number of function keys supported by kernel: %d\n", MAX_NR_FUNC);
-	printf("total space available for compose key definitions: %d bytes\n",
-	        MAX_DIACR);
+	printf("number of function keys supported by kernel: %d\n",
+	       MAX_NR_FUNC);
+
+	printf("max nr of compose definitions: %d\n",
+	       MAX_DIACR);
+#ifdef KDGKBDIACR
+	printf("nr of compose definitions in actual use: %d\n",
+	       nr_of_diacs());
+#endif
 }
 
 struct {
@@ -208,10 +264,17 @@ void dump_symbols(void) {
 	char *p;
 
 	printf("Symbols recognized by dumpkeys:\n(numeric value, symbol)\n\n");
-	for (t = 0; t < syms_size; t++)
+	for (t = 0; t < syms_size; t++) {
+	    if (syms[t].size) {
 		for (v = 0; v < syms[t].size; v++)
 			if ((p = syms[t].table[v])[0])
 				printf("0x%04x\t%s\n", K(t, v), p);
+	    } else if (t == KT_META) {
+		for (v = 0; v < syms[0].size && v < 128; v++)
+			if ((p = syms[0].table[v])[0])
+				printf("0x%04x\tMeta_%s\n", K(t, v), p);
+	    }
+	}
 	printf("\nThe following synonyms are recognized:\n\n");
 	for (t = 0; t < syn_size; t++)
 	  printf("%-15s for %s\n", synonyms[t].synonym,
@@ -242,11 +305,42 @@ void dump_keys(char full_table, char numeric) {
 	int buf[MAX_NR_KEYMAPS];
 	int isletter, islatin, isasexpected;
 	int typ, val;
+	int alt_is_meta = 0;
 
 	get_keymaps();
 	print_keymaps();
 	if (!keymapnr)
 	  return;
+
+	/* first pass: determine whether to set alt_is_meta */
+	for (j = 0; j < MAX_NR_KEYMAPS; j++) {
+	     int ja = (j | (1 << KG_ALT));
+	     if (j != ja && keymap_index[j] >= 0 && keymap_index[ja] >= 0)
+		  for (i = 1; i < NR_KEYS; i++) {
+		       int buf0, buf1, type;
+
+		       buf0 = get_bind(i, j);
+		       type = KTYP(buf0);
+		       if ((type == KT_LATIN || type == KT_LETTER)
+			   && KVAL(buf0) < 128) {
+			    buf1 = get_bind(i, ja);
+			    if (buf1 != K(KT_META, KVAL(buf0))) {
+				 if (verbose) {
+				      printf("# not alt_is_meta: "
+		      		"on keymap %d key %d is bound to",
+					      ja, i);
+				      print_keysym(buf1, numeric);
+				      printf("\n");
+				 }
+				 goto not_alt_is_meta;
+			    }
+		       }
+		  }
+	}
+	alt_is_meta = 1;
+	printf("alt_is_meta\n");
+not_alt_is_meta:
+
 	for (i = 1; i < NR_KEYS; i++) {
 	    for (j = 0; j < keymapnr; j++)
 	      buf[j] = get_bind(i, good_keymap[j]);
@@ -278,6 +372,23 @@ void dump_keys(char full_table, char numeric) {
 	    }
 	  unexpected:
 
+	    /* wipe out predictable meta bindings */
+	    if (alt_is_meta) {
+		 for(j = 0; j < keymapnr; j++) {
+		      int ka, ja, typ;
+		      k = good_keymap[j];
+		      ka = (k | (1 << KG_ALT));
+		      ja = keymap_index[ka];
+		      if (k != ka && ja >= 0
+		       && ((typ=KTYP(buf[j])) == KT_LATIN || typ == KT_LETTER)
+		       && KVAL(buf[j]) < 128) {
+			   if (buf[ja] != K(KT_META, KVAL(buf[j])))
+				fprintf(stderr, "impossible: not meta?\n");
+			   buf[ja] = K_HOLE;
+		      }
+		 }
+	    }
+
 	    printf("keycode %3d =", i);
 	    if (full_table) {
 		for (j = 0; j < keymapnr; j++)
@@ -355,40 +466,6 @@ void dump_funcs(void) {
 	}
 }
 
-#ifdef KDGKBDIACR
-/* isgraph() does not know about iso-8859; printing the character
-   unescaped makes the output easier to check. Maybe this should
-   be an option. Use locale? */
-void
-outchar (unsigned char c) {
-        printf("'");
-        printf((c == '\'' || c == '\\') ? "\\%c"
-	       : (isgraph(c) || c == ' ' || c >= 0200) ? "%c"
-	       : "\\%03o", c);
-	printf("'");
-}
-
-void dump_diacs(void) {
-        struct kbdiacrs kd;
-	int i;
-
-	if(ioctl(fd, KDGKBDIACR, (unsigned long)&kd)) {
-	    fprintf(stderr, "KDGKBDIACR failed\n");
-	    perror("");
-	    exit(1);
-	}
-	for (i = 0; i < kd.kb_cnt; i++) {
-	        printf("compose ");
-		outchar(kd.kbdiacr[i].diacr);
-		printf(" ");
-		outchar(kd.kbdiacr[i].base);
-		printf(" to ");
-		outchar(kd.kbdiacr[i].result);
-		printf("\n");
-	}
-}
-#endif        
-
 void usage(void) {
 	fprintf(stderr, "\
 dumpkeys version " VERSION "\
@@ -405,6 +482,7 @@ valid options are:
 	   --funcs-only	display only the function key strings
 	   --keys-only	display only key bindings
 %s	-c --charset=iso-8859-{1,2,3,4,5,7,8}
+	             koi8-{r,u}		
 			interpret character action codes to be from the
 			specified character set
 ",
@@ -419,7 +497,7 @@ valid options are:
 
 int
 main (int argc, char *argv[]) {
-	const char *short_opts = "hilsnfc:";
+	const char *short_opts = "hilvsnfc:";
 	const struct option long_opts[] = {
 		{ "help",	no_argument,		NULL, 'h' },
 		{ "short-info",	no_argument,		NULL, 'i' },
@@ -431,6 +509,7 @@ main (int argc, char *argv[]) {
 #ifdef KDGKBDIACR
 		{ "compose-only",no_argument,           NULL, 'd' },
 #endif
+		{ "verbose",	no_argument,		NULL, 'v' },
 		{ "charset",	required_argument,	NULL, 'c' },
 		{ NULL,	0, NULL, 0 }
 	};
@@ -444,7 +523,6 @@ main (int argc, char *argv[]) {
 #ifdef KDGKBDIACR
 	char diac_only = 0;
 #endif
-	char charset = 1;
 
 	while ((c = getopt_long(argc, argv,
 		short_opts, long_opts, NULL)) != -1) {
@@ -473,14 +551,13 @@ main (int argc, char *argv[]) {
 				diac_only = 1;
 				break;
 #endif
+			case 'v':
+				verbose = 1;
+				break;
 			case 'c':
-				if (strlen(optarg) != 10 || /* iso-8859-x */
-					strncmp(optarg, "iso-8859-", 9) ||
-					!index("1234578", optarg[9]))
+				if ((set_charset(optarg)) != 0)
 					usage();
-				charset = optarg[9] - '0';
 				printf("charset \"%s\"\n", optarg);
-				set_charset(optarg);
 				break;
 			case 'h':
 			case '?':
