@@ -1,5 +1,7 @@
-#include "ksyms.h"
+#include <linux/keyboard.h>
 #include <stdio.h>
+#include <string.h>
+#include "ksyms.h"
 
 static char *latin1_syms[] = {
 	"nul",
@@ -260,7 +262,7 @@ static char *latin1_syms[] = {
 	"ydiaeresis"
 };
 
-char *latin2_syms[] = {
+static char *latin2_syms[] = {
     "", "Aogonek", "breve", "Lstroke", "", "Lcaron", "Sacute", "",
     "", "Scaron", "Scedilla", "Tcaron", "Zacute", "", "Zcaron", "Zabovedot",
     "", "aogonek", "ogonek", "lstroke", "", "lcaron", "sacute", "caron",
@@ -275,7 +277,27 @@ char *latin2_syms[] = {
     "rcaron", "uring", "", "udoubleacute", "", "", "tcedilla", "abovedot"
 };
 
-char *latin3_syms[] = {
+static char *mazovia_syms[] = { /* as specified by Wlodek Bzyl <matwb@univ.gda.pl> */
+    "", "", "", "", "", "", "aogonek", "", /* 8x */
+    "", "", "", "", "", "cacute", "", "Aogonek",
+    "Eogonek", "eogonek", "lstroke", "", "", "Cacute", "", "", /* 9x */
+    "Sacute", "", "", "", "Lstroke", "", "sacute", "",
+    "Zacute", "Zabovedot", "oacute", "Oacute",
+    "nacute", "Nacute", "zacute", "zabovedot", /* ax */
+    "", "", "", "", "", "", "guillemotleft", "guillemotright",
+    "", "", "", "", "", "", "", "", /* bx */
+    "", "", "", "", "", "", "", "",
+    "", "", "", "", "", "", "", "", /* cx */
+    "", "", "", "", "", "", "", "",
+    "", "", "", "", "", "", "", "", /* dx */
+    "", "", "", "", "", "", "", "",
+    "", "", "", "", "", "", "", "", /* ex */
+    "", "", "", "", "", "", "", "",
+    "", "", "", "", "", "", "", "", /* fx */
+    "", "", "", "", "", "", "", "quotedblbase"
+};
+
+static char *latin3_syms[] = {
     "", "Hstroke", "", "", "", "", "Hcircumflex", "",
     "", "Iabovedot", "", "Gbreve", "Jcircumflex", "", "", "",
     "", "hstroke", "", "", "", "", "hcircumflex", "",
@@ -290,7 +312,7 @@ char *latin3_syms[] = {
     "gcircumflex", "", "", "", "", "ubreve", "scircumflex", ""
 };
 
-char *latin4_syms[] = {
+static char *latin4_syms[] = {
     "", "", "kra", "Rcedilla", "", "Itilde", "Lcedilla", "",
     "", "", "Emacron", "Gcedilla", "Tslash", "", "", "",
     "", "", "", "rcedilla", "", "itilde", "lcedilla", "",
@@ -305,14 +327,14 @@ char *latin4_syms[] = {
     "", "uogonek", "", "", "", "utilde", "umacron", ""
 };
 
-char *iso_8859_5_syms[] = { /* 160-255 */
+static char *iso_8859_5_syms[] = { /* 160-255 */
     "no-break_space",					/* 0240 */
     "cyrillic_capital_letter_io",
     "serbocroatian_cyrillic_capital_letter_dje",
     "macedonian_cyrillic_capital_letter_gje",
     "ukrainian_cyrillic_capital_letter_ie",
     "macedonian_cyrillic_capital_letter_dze",
-    "cyrillic_capital_letter_i",
+    "ukrainian_cyrillic_capital_letter_i",
     "ukrainian_cyrillic_capital_letter_yi",
     "cyrillic_capital_letter_je", 			/* 0250 */
     "cyrillic_capital_letter_lje",
@@ -392,7 +414,7 @@ char *iso_8859_5_syms[] = { /* 160-255 */
     "macedonian_cyrillic_small_letter_gje",
     "ukrainian_cyrillic_small_letter_ie",
     "macedonian_cyrillic_small_letter_dze",
-    "cyrillic_small_letter_i",
+    "ukrainian_cyrillic_small_letter_i",
     "ukrainian_cyrillic_small_letter_yi",
     "cyrillic_small_letter_je",				/* 0370 */
     "cyrillic_small_letter_lje",
@@ -404,7 +426,7 @@ char *iso_8859_5_syms[] = { /* 160-255 */
     "cyrillic_small_letter_dze"
 };
 
-char *iso_8859_7_syms[] = { /* 160-255 */
+static char *iso_8859_7_syms[] = { /* 160-255 */
     "", "leftquote", "rightquote", "", "", "", "", "",
     "", "", "", "", "", "", "", "",
     "", "", "", "", "accent", "diaeresisaccent", "Alphaaccent", "",
@@ -419,7 +441,7 @@ char *iso_8859_7_syms[] = { /* 160-255 */
     "psi", "omega", "iotadiaeresis", "upsilondiaeresis", "omicronaccent", "upsilonaccent", "omegaaccent", ""
 };
 
-char *iso_8859_8_syms[] = {
+static char *iso_8859_8_syms[] = {
     "", "", "", "", "", "", "", "",
     "", "", "multiplication", "", "", "", "", "overscore",
     "", "", "", "", "", "", "", "",
@@ -434,27 +456,110 @@ char *iso_8859_8_syms[] = {
     "resh", "shin", "tav", "", "", "", "", ""
 };
 
+char *koi8_r_syms[] = { /* 160-255 */
+    "",                                        /* 0240 */
+    "",
+    "",
+    "cyrillic_small_letter_io",
+    "",
+    "",
+    "",
+    "",
+    "",                        /* 0250 */
+    "",
+    "",
+    "",
+    "",
+    "",
+    "",
+    "",
+    "",                        /* 0260 */
+    "",
+    "",
+    "cyrillic_capital_letter_io",
+    "",
+    "",
+    "",
+    "",
+    "",                        /* 0270 */
+    "",
+    "",
+    "",
+    "",
+    "",
+    "",
+    "copyright",
+    "cyrillic_small_letter_yu",                        /* 0300 */
+    "cyrillic_small_letter_a",
+    "cyrillic_small_letter_be",
+    "cyrillic_small_letter_tse",
+    "cyrillic_small_letter_de",
+    "cyrillic_small_letter_ie",
+    "cyrillic_small_letter_ef",
+    "cyrillic_small_letter_ghe",
+    "cyrillic_small_letter_ha",                        /* 0310 */
+    "cyrillic_small_letter_i",
+    "cyrillic_small_letter_short_i",
+    "cyrillic_small_letter_ka",
+    "cyrillic_small_letter_el",
+    "cyrillic_small_letter_em",
+    "cyrillic_small_letter_en",
+    "cyrillic_small_letter_o",
+    "cyrillic_small_letter_pe",                                /* 0320 */
+    "cyrillic_small_letter_ya",
+    "cyrillic_small_letter_er",
+    "cyrillic_small_letter_es",
+    "cyrillic_small_letter_te",
+    "cyrillic_small_letter_u",
+    "cyrillic_small_letter_zhe",
+    "cyrillic_small_letter_ve",
+    "cyrillic_small_soft_sign",                                /* 0330 */
+    "cyrillic_small_letter_yeru",
+    "cyrillic_small_letter_ze",
+    "cyrillic_small_letter_sha",
+    "cyrillic_small_letter_e",
+    "cyrillic_small_letter_shcha",
+    "cyrillic_small_letter_che",
+    "cyrillic_small_hard_sign",
+    "cyrillic_capital_letter_yu",                      /* 0340 */
+    "cyrillic_capital_letter_a",
+    "cyrillic_capital_letter_be",
+    "cyrillic_capital_letter_tse",
+    "cyrillic_capital_letter_de",
+    "cyrillic_capital_letter_ie",
+    "cyrillic_capital_letter_ef",
+    "cyrillic_capital_letter_ghe",
+    "cyrillic_capital_letter_ha",                      /* 0350 */
+    "cyrillic_capital_letter_i",
+    "cyrillic_capital_letter_short_i",
+    "cyrillic_capital_letter_ka",
+    "cyrillic_capital_letter_el",
+    "cyrillic_capital_letter_em",
+    "cyrillic_capital_letter_en",
+    "cyrillic_capital_letter_o",
+    "cyrillic_capital_letter_pe",                      /* 0360 */
+    "cyrillic_capital_letter_ya",
+    "cyrillic_capital_letter_er",
+    "cyrillic_capital_letter_es",
+    "cyrillic_capital_letter_te",
+    "cyrillic_capital_letter_u",
+    "cyrillic_capital_letter_zhe",
+    "cyrillic_capital_letter_ve",
+    "cyrillic_capital_soft_sign",                      /* 0370 */
+    "cyrillic_capital_letter_yeru",
+    "cyrillic_capital_letter_ze",
+    "cyrillic_capital_letter_sha",
+    "cyrillic_capital_letter_e",
+    "cyrillic_capital_letter_shcha",
+    "cyrillic_capital_letter_che",
+    "cyrillic_capital_hard_sign"
+};
+
 static char *fn_syms[] = {
-	"F1",
-	"F2",
-	"F3",
-	"F4",
-	"F5",
-	"F6",
-	"F7",
-	"F8",
-	"F9",
-	"F10",
-	"F11",
-	"F12",
-	"F13",
-	"F14",
-	"F15",
-	"F16",
-	"F17",
-	"F18",
-	"F19",
-	"F20",
+	"F1", 	"F2",	"F3",	"F4",	"F5",
+	"F6",	"F7",	"F8",	"F9",	"F10",
+	"F11",	"F12",	"F13",	"F14",	"F15",
+	"F16",	"F17",	"F18",	"F19",	"F20",
 	"Find",			/* also called: "Home" */
 	"Insert",
 	"Remove",
@@ -465,231 +570,51 @@ static char *fn_syms[] = {
 	"Help",
 	"Do",
 	"Pause",
-	"F21",
-	"F22",
-	"F23",
-	"F24",
-	"F25",
-	"F26",
-	"F27",
-	"F28",
-	"F29",
-	"F30",
-	"F31",
-	"F32",
-	"F33",
-	"F34",
-	"F35",
-	"F36",
-	"F37",
-	"F38",
-	"F39",
-	"F40",
-	"F41",
-	"F42",
-	"F43",
-	"F44",
-	"F45",
-	"F46",
-	"F47",
-	"F48",
-	"F49",
-	"F50",
-	"F51",
-	"F52",
-	"F53",
-	"F54",
-	"F55",
-	"F56",
-	"F57",
-	"F58",
-	"F59",
-	"F60",
-	"F61",
-	"F62",
-	"F63",
-	"F64",
-	"F65",
-	"F66",
-	"F67",
-	"F68",
-	"F69",
-	"F70",
-	"F71",
-	"F72",
-	"F73",
-	"F74",
-	"F75",
-	"F76",
-	"F77",
-	"F78",
-	"F79",
-	"F80",
-	"F81",
-	"F82",
-	"F83",
-	"F84",
-	"F85",
-	"F86",
-	"F87",
-	"F88",
-	"F89",
-	"F90",
-	"F91",
-	"F92",
-	"F93",
-	"F94",
-	"F95",
-	"F96",
-	"F97",
-	"F98",
-	"F99",
-	"F100",
-	"F101",
-	"F102",
-	"F103",
-	"F104",
-	"F105",
-	"F106",
-	"F107",
-	"F108",
-	"F109",
-	"F110",
-	"F111",
-	"F112",
-	"F113",
-	"F114",
-	"F115",
-	"F116",
-	"F117",
-	"F118",
-	"F119",
-	"F120",
-	"F121",
-	"F122",
-	"F123",
-	"F124",
-	"F125",
-	"F126",
-	"F127",
-	"F128",
-	"F129",
-	"F130",
-	"F131",
-	"F132",
-	"F133",
-	"F134",
-	"F135",
-	"F136",
-	"F137",
-	"F138",
-	"F139",
-	"F140",
-	"F141",
-	"F142",
-	"F143",
-	"F144",
-	"F145",
-	"F146",
-	"F147",
-	"F148",
-	"F149",
-	"F150",
-	"F151",
-	"F152",
-	"F153",
-	"F154",
-	"F155",
-	"F156",
-	"F157",
-	"F158",
-	"F159",
-	"F160",
-	"F161",
-	"F162",
-	"F163",
-	"F164",
-	"F165",
-	"F166",
-	"F167",
-	"F168",
-	"F169",
-	"F170",
-	"F171",
-	"F172",
-	"F173",
-	"F174",
-	"F175",
-	"F176",
-	"F177",
-	"F178",
-	"F179",
-	"F180",
-	"F181",
-	"F182",
-	"F183",
-	"F184",
-	"F185",
-	"F186",
-	"F187",
-	"F188",
-	"F189",
-	"F190",
-	"F191",
-	"F192",
-	"F193",
-	"F194",
-	"F195",
-	"F196",
-	"F197",
-	"F198",
-	"F199",
-	"F200",
-	"F201",
-	"F202",
-	"F203",
-	"F204",
-	"F205",
-	"F206",
-	"F207",
-	"F208",
-	"F209",
-	"F210",
-	"F211",
-	"F212",
-	"F213",
-	"F214",
-	"F215",
-	"F216",
-	"F217",
-	"F218",
-	"F219",
-	"F220",
-	"F221",
-	"F222",
-	"F223",
-	"F224",
-	"F225",
-	"F226",
-	"F227",
-	"F228",
-	"F229",
-	"F230",
-	"F231",
-	"F232",
-	"F233",
-	"F234",
-	"F235",
-	"F236",
-	"F237",
-	"F238",
-	"F239",
-	"F240",
-	"F241",
-	"F242",
-	"F243",
-	"F244",
-	"F245",
+	"F21",	"F22",	"F23",	"F24",	"F25",
+	"F26",	"F27",	"F28",	"F29",	"F30",
+	"F31",	"F32",	"F33",	"F34",	"F35",
+	"F36",	"F37",	"F38",	"F39",	"F40",
+	"F41",	"F42",	"F43",	"F44",	"F45",
+	"F46",	"F47",	"F48",	"F49",	"F50",
+	"F51",	"F52",	"F53",	"F54",	"F55",
+	"F56",	"F57",	"F58",	"F59",	"F60",
+	"F61",	"F62",	"F63",	"F64",	"F65",
+	"F66",	"F67",	"F68",	"F69",	"F70",
+	"F71",	"F72",	"F73",	"F74",	"F75",
+	"F76",	"F77",	"F78",	"F79",	"F80",
+	"F81",	"F82",	"F83",	"F84",	"F85",
+	"F86",	"F87",	"F88",	"F89",	"F90",
+	"F91",	"F92",	"F93",	"F94",	"F95",
+	"F96",	"F97",	"F98",	"F99",	"F100",
+	"F101",	"F102",	"F103",	"F104",	"F105",
+	"F106",	"F107",	"F108",	"F109",	"F110",
+	"F111",	"F112",	"F113",	"F114",	"F115",
+	"F116",	"F117",	"F118",	"F119",	"F120",
+	"F121",	"F122",	"F123",	"F124",	"F125",
+	"F126",	"F127",	"F128",	"F129",	"F130",
+	"F131",	"F132",	"F133",	"F134",	"F135",
+	"F136",	"F137",	"F138",	"F139",	"F140",
+	"F141",	"F142",	"F143",	"F144",	"F145",
+	"F146",	"F147",	"F148",	"F149",	"F150",
+	"F151",	"F152",	"F153",	"F154",	"F155",
+	"F156",	"F157",	"F158",	"F159",	"F160",
+	"F161",	"F162",	"F163",	"F164",	"F165",
+	"F166",	"F167",	"F168",	"F169",	"F170",
+	"F171",	"F172",	"F173",	"F174",	"F175",
+	"F176",	"F177",	"F178",	"F179",	"F180",
+	"F181",	"F182",	"F183",	"F184",	"F185",
+	"F186",	"F187",	"F188",	"F189",	"F190",
+	"F191",	"F192",	"F193",	"F194",	"F195",
+	"F196",	"F197",	"F198",	"F199",	"F200",
+	"F201",	"F202",	"F203",	"F204",	"F205",
+	"F206",	"F207",	"F208",	"F209",	"F210",
+	"F211",	"F212",	"F213",	"F214",	"F215",
+	"F216",	"F217",	"F218",	"F219",	"F220",
+	"F221",	"F222",	"F223",	"F224",	"F225",
+	"F226",	"F227",	"F228",	"F229",	"F230",
+	"F231",	"F232",	"F233",	"F234",	"F235",
+	"F236",	"F237",	"F238",	"F239",	"F240",
+	"F241",	"F242",	"F243",	"F244",	"F245",
 	"F246"		/* there are 10 keys named Insert etc., total 256 */
 };
 
@@ -742,7 +667,8 @@ static char *dead_syms[] = {
 	"dead_acute",
 	"dead_circumflex",
 	"dead_tilde",
-	"dead_diaeresis"
+	"dead_diaeresis",
+	"dead_cedilla"
 };
 
 static char *cons_syms[] = {
@@ -830,137 +756,6 @@ static char *shift_syms[] = {
 	"CapsShift"
 };
 
-static char *meta_syms[] = {
-	"Meta_nul",
-	"Meta_Control_a",
-	"Meta_Control_b",
-	"Meta_Control_c",
-	"Meta_Control_d",
-	"Meta_Control_e",
-	"Meta_Control_f",
-	"Meta_Control_g",
-	"Meta_BackSpace",
-	"Meta_Tab",
-	"Meta_Linefeed",
-	"Meta_Control_k",
-	"Meta_Control_l",
-	"Meta_Control_m",
-	"Meta_Control_n",
-	"Meta_Control_o",
-	"Meta_Control_p",
-	"Meta_Control_q",
-	"Meta_Control_r",
-	"Meta_Control_s",
-	"Meta_Control_t",
-	"Meta_Control_u",
-	"Meta_Control_v",
-	"Meta_Control_w",
-	"Meta_Control_x",
-	"Meta_Control_y",
-	"Meta_Control_z",
-	"Meta_Escape",
-	"Meta_Control_backslash",
-	"Meta_Control_bracketright",
-	"Meta_Control_asciicircum",
-	"Meta_Control_underscore",
-	"Meta_space",
-	"Meta_exclam",
-	"Meta_quotedbl",
-	"Meta_numbersign",
-	"Meta_dollar",
-	"Meta_percent",
-	"Meta_ampersand",
-	"Meta_apostrophe",
-	"Meta_parenleft",
-	"Meta_parenright",
-	"Meta_asterisk",
-	"Meta_plus",
-	"Meta_comma",
-	"Meta_minus",
-	"Meta_period",
-	"Meta_slash",
-	"Meta_zero",
-	"Meta_one",
-	"Meta_two",
-	"Meta_three",
-	"Meta_four",
-	"Meta_five",
-	"Meta_six",
-	"Meta_seven",
-	"Meta_eight",
-	"Meta_nine",
-	"Meta_colon",
-	"Meta_semicolon",
-	"Meta_less",
-	"Meta_equal",
-	"Meta_greater",
-	"Meta_question",
-	"Meta_at",
-	"Meta_A",
-	"Meta_B",
-	"Meta_C",
-	"Meta_D",
-	"Meta_E",
-	"Meta_F",
-	"Meta_G",
-	"Meta_H",
-	"Meta_I",
-	"Meta_J",
-	"Meta_K",
-	"Meta_L",
-	"Meta_M",
-	"Meta_N",
-	"Meta_O",
-	"Meta_P",
-	"Meta_Q",
-	"Meta_R",
-	"Meta_S",
-	"Meta_T",
-	"Meta_U",
-	"Meta_V",
-	"Meta_W",
-	"Meta_X",
-	"Meta_Y",
-	"Meta_Z",
-	"Meta_bracketleft",
-	"Meta_backslash",
-	"Meta_bracketright",
-	"Meta_asciicircum",
-	"Meta_underscore",
-	"Meta_grave",
-	"Meta_a",
-	"Meta_b",
-	"Meta_c",
-	"Meta_d",
-	"Meta_e",
-	"Meta_f",
-	"Meta_g",
-	"Meta_h",
-	"Meta_i",
-	"Meta_j",
-	"Meta_k",
-	"Meta_l",
-	"Meta_m",
-	"Meta_n",
-	"Meta_o",
-	"Meta_p",
-	"Meta_q",
-	"Meta_r",
-	"Meta_s",
-	"Meta_t",
-	"Meta_u",
-	"Meta_v",
-	"Meta_w",
-	"Meta_x",
-	"Meta_y",
-	"Meta_z",
-	"Meta_braceleft",
-	"Meta_bar",
-	"Meta_braceright",
-	"Meta_asciitilde",
-	"Meta_Delete"
-};
-
 static char *ascii_syms[] = {
 	"Ascii_0",
 	"Ascii_1",
@@ -1001,6 +796,17 @@ static char *lock_syms[] = {
 	"CtrlR_Lock"
 };
 
+static char *sticky_syms[] = {
+	"SShift",
+	"SAltGr",
+	"SControl",
+	"SAlt",
+	"SShiftL",
+	"SShiftR",
+	"SCtrlL",
+	"SCtrlR"
+};
+
 #define E(x) { x, sizeof(x) / sizeof(char **) }
 
 syms_entry syms[] = {
@@ -1012,9 +818,11 @@ syms_entry syms[] = {
 	E(cons_syms),
 	E(cur_syms),
 	E(shift_syms),
-	E(meta_syms),
+	{ 0, 0 },		/* KT_META */
 	E(ascii_syms),
-	E(lock_syms)   
+	E(lock_syms),
+	{ 0, 0 },		/* KT_LETTER */
+	E(sticky_syms)
 };
 
 #undef E
@@ -1024,9 +832,6 @@ synonyms[] = {
 	{ "Control_h",		"BackSpace" },
 	{ "Control_i",		"Tab" },
 	{ "Control_j",		"Linefeed" },
-	{ "Meta_Control_h",	"Meta_BackSpace" },
-	{ "Meta_Control_i",	"Meta_Tab" },
-	{ "Meta_Control_j",	"Meta_Linefeed" },
 	{ "Home",		"Find" },
 /* Unfortunately Delete already denotes ASCII 0177 */
 /*	{ "Delete",		"Remove" }, */
@@ -1049,6 +854,7 @@ synonyms[] = {
 	{ "AltGr_R",		"AltGr" },
 	{ "AltLLock",		"Alt_Lock" },
 	{ "AltRLock",		"AltGr_Lock" },
+	{ "SCtrl",		"SControl" },
 	{ "KeyboardSignal",	"Spawn_Console" },
 /* the names of the Greek letters are spelled differently
    in the iso-8859-7 and the Unicode standards */
@@ -1061,43 +867,126 @@ synonyms[] = {
 /* diacriticals */
 	{ "tilde",		"asciitilde" },
 	{ "circumflex",		"asciicircum" },
+/* as dead_ogonek, dead_caron, dead_breve and dead_doubleacute doesn't exist
+ * (yet), I put also compose lines for use with respectively dead_cedilla,
+ * dead_circumflex, dead_tilde and dead_tilde */
+	{ "dead_ogonek",        "dead_cedilla" },
+	{ "dead_caron",         "dead_circumflex" },
+	{ "dead_breve",         "dead_tilde" },
+	{ "dead_doubleacute",   "dead_tilde" },
+/* cyrillic */
+	{ "bielorussian_cyrillic_capital_letter_i", "ukrainian_cyrillic_capital_letter_i" },
+	{ "cyrillic_capital_letter_kha", "cyrillic_capital_letter_ha" },
+	{ "cyrillic_capital_letter_ge", "cyrillic_capital_letter_ghe" },
+	{ "cyrillic_capital_letter_ia", "cyrillic_capital_letter_ya" },
+	{ "cyrillic_capital_letter_iu", "cyrillic_capital_letter_yu" },
+	{ "cyrillic_capital_letter_yeri", "cyrillic_capital_letter_yeru" },
+	{ "cyrillic_capital_letter_reversed_e", "cyrillic_capital_letter_e" },
+	{ "cyrillic_capital_letter_ii", "cyrillic_capital_letter_i" },
+	{ "cyrillic_capital_letter_short_ii", "cyrillic_capital_letter_short_i" },
+	{ "bielorussian_cyrillic_small_letter_i", "ukrainian_cyrillic_small_letter_i" },
+	{ "cyrillic_small_letter_kha", "cyrillic_small_letter_ha" },
+	{ "cyrillic_small_letter_ge", "cyrillic_small_letter_ghe" },
+	{ "cyrillic_small_letter_ia", "cyrillic_small_letter_ya" },
+	{ "cyrillic_small_letter_iu", "cyrillic_small_letter_yu" },
+	{ "cyrillic_small_letter_yeri", "cyrillic_small_letter_yeru" },
+	{ "cyrillic_small_letter_reversed_e", "cyrillic_small_letter_e" },
+	{ "cyrillic_small_letter_ii", "cyrillic_small_letter_i" },
+	{ "cyrillic_small_letter_short_ii", "cyrillic_small_letter_short_i" }
 };
 
 const int syms_size = sizeof(syms) / sizeof(syms_entry);
 const int syn_size = sizeof(synonyms) / sizeof(synonyms[0]);
 
-/* this is silly, I'll fix it some day */
 struct cs {
     char *charset;
     char **charnames;
+    int start;
 } charsets[] = {
-    { "iso-8859-1", 0 },
-    { "iso-8859-2", latin2_syms },
-    { "iso-8859-3", latin3_syms },
-    { "iso-8859-4", latin4_syms },
-    { "iso-8859-5", iso_8859_5_syms },
-    { "iso-8859-7", iso_8859_7_syms },
-    { "iso-8859-8", iso_8859_8_syms }
+    { "iso-8859-1", 0, 0 },
+    { "mazovia", mazovia_syms, 128 },
+    { "iso-8859-2", latin2_syms, 160 },
+    { "iso-8859-3", latin3_syms, 160 },
+    { "iso-8859-4", latin4_syms, 160 },
+    { "iso-8859-5", iso_8859_5_syms, 160 },
+    { "iso-8859-7", iso_8859_7_syms, 160 },
+    { "iso-8859-8", iso_8859_8_syms, 160 },
+    { "koi8-r",     koi8_r_syms, 160 },
+    { "koi8-u",     koi8_r_syms, 160 }
+/* I don't know the differences between koi8-r and koi8-u,
+   so I use the same syms table - Pablo Saratxaga */
 };
 
-void
+int
 set_charset(char *charset) {
 	char **p;
 	int i;
 
 	for (i = 0; i < sizeof(charsets)/sizeof(charsets[0]); i++) {
-	    if (!strcmp(charsets[i].charset, charset)) {
+	    if (!strcasecmp(charsets[i].charset, charset)) {
 		p = charsets[i].charnames;
 		if (p)
-		  for (i = 160; i < 256; i++,p++) {
+		  for (i = charsets[i].start; i < 256; i++,p++) {
 		    if(*p && **p)
 		      syms[0].table[i] = *p;
 		}
-		return;
+		return 0;
 	    }
 	}
-	if (!strcmp(charset, "unicode"))
+	if (!strcasecmp(charset, "unicode")) {
 	  fprintf (stderr, "unknown charset %s - ignoring charset request\n",
 		   charset);
+	  return 1;
+	}
+	return 0;
 }
 
+int
+ksymtocode(char *s)
+{
+	int i;
+	int j;
+	int keycode;
+	static warn = 0;
+
+	if (!strncmp(s, "Meta_", 5)) {
+		keycode = ksymtocode(s+5);
+		if (KTYP(keycode) == KT_LATIN)
+			return K(KT_META, KVAL(keycode));
+		/* fall through to error printf */
+	}
+
+	for (i = 0; i < syms_size; i++)
+		for (j = 0; j < syms[i].size; j++)
+			if (!strcmp(s,syms[i].table[j]))
+				return K(i, j);
+
+	for (i = 0; i < syn_size; i++)
+		if (!strcmp(s, synonyms[i].synonym))
+			return ksymtocode(synonyms[i].official_name);
+
+	for (i = 0; i < 256 - 160; i++)
+		if (!strcmp(s, latin2_syms[i])) {
+			if(!warn++)
+				fprintf(stderr, "assuming iso-8859-2\n");
+			return K(KT_LATIN, 160 + i);
+		}
+
+	for (i = 0; i < 256 - 160; i++)
+		if (!strcmp(s, latin3_syms[i])) {
+			if(!warn++)
+				fprintf(stderr, "assuming iso-8859-3\n");
+			return K(KT_LATIN, 160 + i);
+		}
+
+	for (i = 0; i < 256 - 160; i++)
+		if (!strcmp(s, latin4_syms[i])) {
+			if(!warn++)
+				fprintf(stderr, "assuming iso-8859-4\n");
+			return K(KT_LATIN, 160 + i);
+		}
+
+	fprintf(stderr, "unknown keysym '%s'\n", s);
+
+	return -1;
+}
