@@ -76,9 +76,13 @@
 #include <errno.h>
 #include <signal.h>
 #include <sys/ioctl.h>
+#if (__GNU_LIBRARY__ >= 6)
+#include <sys/perm.h>
+#else
 #include <linux/types.h>
-#include <linux/vt.h>
 #include <linux/termios.h>
+#endif
+#include <linux/vt.h>
 #include "paths.h"
 
 #define MODE_RESTORETEXTMODE	0
@@ -136,7 +140,7 @@ main(argc, argv) int argc; char **argv; {
 	    fprintf(stderr, "resizecons: cannot find videomode file %s\n", infile);
 	    exit(1);
 	}
- 	fclose(fin);
+ 	fpclose(fin);
     }
 
     fd = getfd();
