@@ -1,7 +1,7 @@
 /*
  * setfont.c - Eugene Crosser & Andries Brouwer
  *
- * Version 0.92
+ * Version 0.94
  *
  * Loads the console font, and possibly the corresponding screen map(s).
  * We accept two kind of screen maps, one [-m] giving the correspondence
@@ -9,7 +9,7 @@
  * font positions, and the second [-u] giving the correspondence between
  * font positions and Unicode values.
  */
-#define VERSION "0.92"
+#define VERSION "0.94"
 
 #include <stdio.h>
 #include <memory.h>
@@ -175,8 +175,7 @@ do_loadfont(int fd, int unit, int fontsize, FILE *fpi) {
 
 	if (verbose)
 	  printf("Loading 8x%d font from file %s\n", unit, pathname);
-
-#ifdef PIO_FONTX
+#if defined( PIO_FONTX ) && !defined( sparc )
 	{
 	    struct consolefontdesc cfd;
 
@@ -227,6 +226,11 @@ do_loadtable(int fd, int tailsz, int fontsize, FILE *fpi) {
 	    }
 	}
 
+	/* Note: after PIO_UNIMAPCLR and before PIO_UNIMAP
+	   this printf does not work. */
+	if (verbose)
+	  printf("Loading Unicode mapping table...\n");
+
 	advice.advised_hashsize = 0;
 	advice.advised_hashstep = 0;
 	advice.advised_hashlevel = 0;
@@ -240,8 +244,6 @@ do_loadtable(int fd, int tailsz, int fontsize, FILE *fpi) {
 	      perror("PIO_UNIMAPCLR");
 	    exit(1);
 	}
-	if (verbose)
-	  printf("Loading Unicode mapping table...\n");
 	ud.entry_ct = ct;
 	ud.entries = up;
 	if(ioctl(fd, PIO_UNIMAP, &ud)) {
@@ -306,7 +308,7 @@ loadnewfont(int fd, char *ifil, int iunit, int no_m, int no_u) {
 		perror("Error reading header input font");
 		exit(1);
 	    }
-	    if (psfhdr.magic != PSF_MAGIC) {
+	    if (!PSF_MAGIC_OK(psfhdr)) {
 		fseek(fpi,0,SEEK_SET);
 		goto no_psf;
 	    }
@@ -315,7 +317,7 @@ loadnewfont(int fd, char *ifil, int iunit, int no_m, int no_u) {
 		exit(1);
 	    }
 	    fontsize = ((psfhdr.mode & PSF_MODE512) ? 512 : 256);
-#ifndef PIO_FONTX
+#if !defined( PIO_FONTX ) || defined( sparc )
 	    if (fontsize != 256) {
 		fprintf(stderr, "Only fontsize 256 supported\n");
 		exit(1);
@@ -331,7 +333,7 @@ loadnewfont(int fd, char *ifil, int iunit, int no_m, int no_u) {
 	    do_loadfont(fd, unit, fontsize, fpi);
 	    if (hastable && !no_u)
 	      do_loadtable(fd, size-head, fontsize, fpi);
-	    fclose(fpi);
+	    fpclose(fpi);
 	    return;
 	}
       no_psf:
@@ -349,7 +351,7 @@ loadnewfont(int fd, char *ifil, int iunit, int no_m, int no_u) {
 	    unit = size/256;
 	}
 	do_loadfont(fd, unit, 256, fpi);
-	fclose(fpi);
+	fpclose(fpi);
 }
 
 static void
